local MAJOR, MINOR = "LibFail-1.0", tonumber("50") or 999 

assert(LibStub, MAJOR.." requires LibStub")

local lib = LibStub:NewLibrary(MAJOR, MINOR)
if not lib then return end

local callbacks = LibStub("CallbackHandler-1.0"):New(lib)

lib.frame = lib.frame or CreateFrame("Frame")

local frame = lib.frame
frame:RegisterEvent("PLAYER_ENTERING_WORLD")


--- Fail Events.
--@description The list of supported events
--@class table
--@name fail_events
--@field 1 Fail_Sartharion_LavaWaves
--@field 2 Fail_Heigan_Dance
--@field 3 Fail_Sartharion_VoidZone
--@field 4 Fail_KelThuzad_VoidZone
--@field 5 Fail_Sapphiron_Breath
--@field 6 Fail_Frogger
--@field 7 Fail_Thaddius_PolaritySwitch
--@field 8 Fail_Thaddius_Jump
--@field 9 Fail_Mimiron_Rocket
--@field 10 Fail_Deconstructor_Light
--@field 11 Fail_Deconstructor_Void
local fail_events = {
    "Fail_Sartharion_LavaWaves",
    "Fail_Heigan_Dance",
    "Fail_Sartharion_VoidZone",
    "Fail_KelThuzad_VoidZone",
    "Fail_Sapphiron_Breath",
    "Fail_Frogger",
    "Fail_Thaddius_PolaritySwitch",
    "Fail_Thaddius_Jump",
    "Fail_Deconstructor_Light",
    "Fail_Deconstructor_Void",
    "Fail_Mimiron_Rocket",
    "Fail_Thorim_LightningCharge",
    "Fail_Thorim_LightningChain",
    "Fail_Vezax_ShadowCrash",
    "Fail_Mimiron_Shock",
    "Fail_IronCouncil_Smash",
    "Fail_Mimiron_BombBots"
}

--[===[@debug@ 
function lib:Test(overrideName)
    local e = math.floor(math.random() * #fail_events) + 1
    local p = math.floor(math.random() * 5) + 1

    self:FailEvent(fail_events[e], overrideName or "Test"..p)
end
--@end-debug@]===]

-- Not sure if this really gets used
local survivable_fails = {
    Fail_Sartharion_LavaWaves = true,
    Fail_Sartharion_VoidZone = true,
    Fail_Heigan_Dance = true,
    Fail_Thaddius_PolaritySwitch = true,
    Fail_Thaddius_Jump = true,
    Fail_Deconstructor_Light = true,
    Fail_Deconstructor_Void = true,
    Fail_Vezax_ShadowCrash = true
}

-- The distinction between the mass/group is there is 1 fail 
-- for N people in a group fail. For mass fails, there are N fails
-- for N people.

-- NYI, fails which have a group component
-- that is *part* of the fail 
-- eg. chain lightning
local group_fails = {
    Fail_Thorim_LightningChain = true,
}

-- NYI, fails which tend to happen to alot of
-- people at the same time - but are still individual fails
-- eg. lava waves (this info may be redundant)
--local mass_fails = {
--    Fail_Sartharion_LavaWaves = true,
--    Fail_Heigan_Dance = true,
--    Fail_Sartharion_VoidZone = true,
--}

local zones_with_fails = {
    Ulduar = {
        "Fail_Deconstructor_Light",
        "Fail_Deconstructor_Void",
        "Fail_Mimiron_Rocket",
        "Fail_Thorim_LightningCharge",
        "Fail_Thorim_LightningChain",
        "Fail_Mimiron_Shock",
        "Fail_IronCouncil_Smash",
        "Fail_Vezax_ShadowCrash",
        "Fail_Mimiron_BombBots",
    },
    Naxxramas = {
        "Fail_Heigan_Dance",
        "Fail_Thaddius_PolaritySwitch",
        "Fail_Thaddius_Jump",
        "Fail_Frogger",
        "Fail_KelThuzad_VoidZone",
        "Fail_Sapphiron_Breath"
    }, 
    ["The Obsidian Sanctum"] = {
        "Fail_Sartharion_LavaWaves",
        "Fail_Sartharion_VoidZone"
    }
}

--- Get a list of supported events.
-- @see fail_events
-- @return a table of event names which can be fired
function lib:GetSupportedEvents() return fail_events end

--- Get a list of supported events in the current zone
-- @see fail_events
-- @return a table of event names which can be fired
function lib:GetSupportedZoneEvents(name) return zones_with_fails[name] end


--- Is the event survivable.
--@param event an event from the list of supported events
--@return true or nil
function lib:IsSurvivable(event) return survivable_fails[event] end

--- Is the event a failure by a group or an individual. Group failures
-- will come as several events in quick sucession, this gives the host a hint
-- as to how to group and display the events.
--@param event an event from the list of supported events
--@return true or nil
function lib:IsGroupFail(event) return group_fails[event] end


function lib:GoInactive()
    if not self.active then return end

    self.SquelchFail = {}
    self.FailThrottle = {}
    self.LastTsunami = {}
    self.ChargeCounter = {}
    self.LastCharge = {}
    self.ThaddiusAlive = true;
    self.DeathTime = 0
    self.LastSlime = {}
    self.FroggerTime = {}

    self.LastGravity, self.LastLight = {}, {}
    self.active = nil

    frame:RegisterEvent("RAID_ROSTER_UPDATE")
    frame:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED")

    callbacks:Fire("Fail_Inactive")
end

lib.active = true
lib:GoInactive()

lib.THADDIUS_JUMP_WINDOW = 120
lib.THADDIUS_JUMP_RETRY_WINDOW = 5
lib.FROGGER_DEATH_WINDOW = 4

do
    local _, etype, f
    
    frame:SetScript("OnEvent", function (self, event, ...)
        if event == "COMBAT_LOG_EVENT_UNFILTERED" then
            _, etype = ...
            f = lib[etype]
    
            if f then
                f(lib, ...)
            end
            
            -- This needs to be moved elsewhere
            local timestamp, type, sourceGUID, sourceName, sourceFlags, destGUID, destName = ...
            if destName then 
                if (destName=="Stalagg") or (destName=="Feugen") then
                    self.DeathTime = timestamp
                end
            end
            
            return
        end

        f = lib[event]

        if f then
            f(lib, ...)
        end
    end)
end


function lib:FailEvent(name, playername)
    callbacks:Fire(name, playername, survivable_fails[name])
end


function lib:GoActive()
    if self.active then return end

    --frame:UnregisterEvent("RAID_ROSTER_UPDATE")
    frame:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")

    self.active = true

    callbacks:Fire("Fail_Active")
end


function lib:PLAYER_ENTERING_WORLD(...)
    if GetNumRaidMembers() > 0 then
        self:GoActive()
    else
        self:GoInactive()
    end
end

function lib:RAID_ROSTER_UPDATE(...)
    if GetNumRaidMembers() > 0 then
        self:GoActive()
    else
        self:GoInactive()
    end
end


function lib:SPELL_DAMAGE(timestamp, type, sourceGUID, sourceName, sourceFlags, destGUID, destName, destFlags, spellId, spellName, spellSchool)



    -- Guardian activities ignored after this point
--5/7 21:13:33.865  SPELL_DAMAGE,0xF1300079F0003A98,"Mirror Image",0x2114,0xF130008092003842,"Elder Stonebark",0xa48,59637,"Fire Blast",0x4,139,0,4,0,0,0,nil,nil,nil
--5/7 21:13:36.092  SPELL_HEAL,0x01800000007C56B2,"Blackknite",0x512,0xF1300079F0003A97,"Mirror Image",0x2114,54968,"Glyph of Holy Light",0x2,1240,1240,nil
    if (bit.band(sourceFlags,COMBATLOG_OBJECT_TYPE_GUARDIAN)>0) or (bit.band(destFlags,COMBATLOG_OBJECT_TYPE_GUARDIAN)>0) then
        return  
    end

    -- Lava Waves: Flame Tsunami 57491
    if spellId == 57491 then
        if bit.band(destFlags, COMBATLOG_OBJECT_TYPE_PLAYER) > 0 then
            if self.LastTsunami[destName] == nil then
                self.LastTsunami[destName] = timestamp
                
                self:FailEvent("Fail_Sartharion_LavaWaves", destName)
            else
                if (timestamp - lib.LastTsunami[destName]) > 10 then
                    self:FailEvent("Fail_Sartharion_LavaWaves", destName)
                end
                self.LastTsunami[destName] = timestamp
            end
        end

        return
    end

    -- Deconstructor: Light Bomb
    if spellId == 65120 and sourceGUID ~= destGUID then
        if bit.band(destFlags, COMBATLOG_OBJECT_TYPE_PLAYER) > 0 then
            if self.LastLight[sourceName] ~= nil then
                if (timestamp - lib.LastLight[sourceName]) > 3 then
                    self:FailEvent("Fail_Deconstructor_Light", sourceName)
                end
                self.LastLight[sourceName] = nil
            end
        end

        return
    end

    -- Deconstructor: Gravity Bomb (bomb part)
    if spellId == 64233 and sourceGUID ~= destGUID then
        if bit.band(destFlags, COMBATLOG_OBJECT_TYPE_PLAYER) > 0 then
            if self.LastGravity[sourceName] ~= nil then
                self:FailEvent("Fail_Deconstructor_Void", sourceName)
                self.LastGravity[sourceName] = nil
            end
        end

        return
    end

-- Failbot - Viiv
--4/16 18:50:56.578  SPELL_DAMAGE,0x0000000000000000,nil,0x80000000,0x05000000027ECB89,"Logicalness",0x514,64875,"Sapper Explosion",0x40,67542,45099,64,28421,0,0,nil,nil,nil

    if spellId == 64875 then
        if bit.band(destFlags, COMBATLOG_OBJECT_TYPE_PLAYER) > 0 then
            FailBot_AddFail(destName,"Fail_Boss_Sapper")
        end
        return
    end


-- Failbot - Viiv
--4/16 21:10:24.039  SPELL_DAMAGE,0xF1300083EC02D966,"Faceless Horror",0x10a48,0x05000000027ECC71,"Boniface",0x514,63721,"Shadow Crash",0x20,8103,0,32,2457,0,0,nil,nil,nil
--4/16 22:06:17.885  SPELL_DAMAGE,0xF1300081F702D928,"General Vezax",0x10a48,0x05000000027FCDFE,"Kosie",0x514,62659,"Shadow Crash",0x20,9413,0,32,2285,0,0,nil,nil,nil
    if spellId == 62659 or spellId == 63721 then
        if bit.band(destFlags, COMBATLOG_OBJECT_TYPE_PLAYER) > 0 then
            self:FailEvent("Fail_Vezax_ShadowCrash", destName)
        end
        return
    end

-- Failbot - Viiv
--4/16 18:20:24.295  SPELL_DAMAGE,0xF130008061018374,"Thorim",0x8010a48,0x0500000001E8AF39,"Thefeint",0x514,62466,"Lightning Charge",0x8,8977,0,8,3966,0,0,nil,nil,nil
    if spellId == 62466 --[[and sourceName=="Thorim"]] then
        if bit.band(destFlags, COMBATLOG_OBJECT_TYPE_PLAYER) > 0 then
            self:FailEvent("Fail_Thorim_LightningCharge", destName)
        end
        return
    end

    -- Chain Lightning
    if spellId == 64390 then
        -- The event starts from the end of the chain and works back to the start. Who failed?
        -- Need a way of expressing group failures
        if bit.band(destFlags, COMBATLOG_OBJECT_TYPE_PLAYER) > 0 then
            self:FailEvent("Fail_Thorim_LightningChain", destName)
        end
        return
    end

    -- Rocket Strike
    if spellId == 63041 then
        if bit.band(destFlags, COMBATLOG_OBJECT_TYPE_PLAYER) > 0 then
            self:FailEvent("Fail_Mimiron_Rocket", destName)
        end
        return
    end

-- Failbot - Viiv
--4/16 13:35:12.750  SPELL_DAMAGE,0xF13000842C0094E3,"Bomb Bot",0xa48,0x05000000027ECCA5,"Naddia",0x512,63801,"Bomb Bot",0x4,20216,4025,4,5054,0,0,nil,nil,nil
    if spellId == 63801 then
        if timestamp - self.SquelchFail.Fail_Mimiron_BombBots <= 0 then
            return
        end
        if bit.band(destFlags, COMBATLOG_OBJECT_TYPE_PLAYER) > 0 then
            self:FailEvent("Fail_Mimiron_BombBots", destName)
        end
        return
    end

--4/16 01:06:26.414  SPELL_DAMAGE,0x0000000000000000,nil,0x80000000,0x05000000027ECA9C,"Cn",0x514,62465,"Runic Smash",0x4,6544,0,4,3116,0,0,nil,nil,nil
    if spellId == 62465 then
        if bit.band(destFlags, COMBATLOG_OBJECT_TYPE_PLAYER) > 0 then
            self:FailEvent("Fail_IronCouncil_Smash", destName)
        end
        return
    end

--3/13 21:17:23.756  SPELL_DAMAGE,0xF150008298002210,"Leviathan Mk II",0x10a48,0xF1300007AC0025A9,"Treant",0x1114,63631,"Shock Blast",0x8,97000,92908,8,0,0,0,nil,nil,nil
    if spellId == 63631 then
        if bit.band(destFlags, COMBATLOG_OBJECT_TYPE_PLAYER) > 0 then
            self:FailEvent("Fail_Mimiron_Shock", destName)
        end
        return
    end



    -- Heigan Dance, Void Zones, Sapphiron's Breath, Frogger

    -- Frogger: Living Poision -> Explode 28433
    -- Heigan: Eruption 29371
    -- Sapphiron: Frost Breath 28524
    -- Void Blast 27812
    local FailType
--  if (spellName=="Eruption") or (spellName=="Void Blast") or (spellName=="Frost Breath") or (spellName=="Explode") then
--      if (sourceName:find("Fissure")) or (sourceName=="Sapphiron") or (sourceName=="Living Poison") then


    if spellId == 29371 then
        FailType = "Heigan_Dance"
    elseif (spellId == 57581) or (spellId == 59128) then
        FailType = "Sartharion_VoidZone"
    elseif spellId == 27812 then
        FailType = "KelThuzad_VoidZone"
    elseif spellId == 28524 then
        FailType = "Sapphiron_Breath"
    elseif spellId == 28433 then
        self.FroggerTime[destName] = timestamp
        return
    end


    if FailType and bit.band(destFlags, COMBATLOG_OBJECT_TYPE_PLAYER) > 0 then
        self:FailEvent("Fail_"..FailType, destName)
        return
    end

--      end
--  end

    -- Thaddius Polarity Switch
    -- 28062 Positive Charge
    -- 28085 Negative Charge
    if self.ThaddiusAlive then
        if spellId == 28062  or spellId == 28085 then
            if self.ChargeCounter[sourceName] == nil then
                self.ChargeCounter[sourceName] = 1
                self.LastCharge[sourceName] = timestamp
            elseif (timestamp - lib.LastCharge[sourceName]) < 2 then
                self.ChargeCounter[sourceName] = lib.ChargeCounter[sourceName] + 1
                self.LastCharge[sourceName] = timestamp
            else
                self.ChargeCounter[sourceName] = 1
                self.LastCharge[sourceName] = timestamp
            end

            if self.ChargeCounter[sourceName] == 3 then
                callbacks:Fire("Fail_Thaddius_PolaritySwitch", sourceName)
            end
        end
    end
end


function lib:ENVIRONMENTAL_DAMAGE(timestamp, type, sourceGUID, sourceName, sourceFlags, destGUID, destName, destFlags, dmgType)
    if (timestamp - self.DeathTime) < self.THADDIUS_JUMP_WINDOW and dmgType == "FALLING" then 
        if bit.band(destFlags, COMBATLOG_OBJECT_TYPE_PLAYER) > 0 then    
            self.LastSlime[destName] = timestamp
            self:FailEvent("Fail_Thaddius_Jump", destName)
        end
    end
end

-- 28089 Polarity Shift 
function lib:SPELL_CAST_START(timestamp, type, sourceGUID, sourceName, sourceFlags, destGUID, destName, destFlags, spellId, spellName, spellSchool)
--  if sourceName=="Thaddius" then
        if spellId == 28089 then
            wipe(self.ChargeCounter)
            self.ThaddiusAlive = true;
        end
--  end
end

function lib:UNIT_DIED(timestamp, type, sourceGUID, sourceName, sourceFlags, destGUID, destName, destFlags)
    if (destName=="Thaddius") then
        self.ThaddiusAlive = false
    elseif self.FroggerTime[destName] then

        if (timestamp - self.FroggerTime[destName] < self.FROGGER_DEATH_WINDOW ) and bit.band(destFlags, COMBATLOG_OBJECT_TYPE_PLAYER) > 0  then
           self:FailEvent("Fail_Frogger", destName)
        end

        self.FroggerTime[destName] = nil
    end


    
end

-- 28801 Slime (Under Thaddius)
function lib:SPELL_AURA_APPLIED(timestamp, type, sourceGUID, sourceName, sourceFlags, destGUID, destName, destFlags, spellId, spellName, spellSchool)
    if spellId == 28801 then
        if (timestamp - self.DeathTime) < self.THADDIUS_JUMP_WINDOW then -- error
            if bit.band(destFlags, COMBATLOG_OBJECT_TYPE_PLAYER) > 0 then
                if self.LastSlime[destName] == nil then
                    self:FailEvent("Fail_Thaddius_Jump", destName)
                elseif (timestamp - self.LastSlime[destName]) > self.THADDIUS_JUMP_RETRY_WINDOW then
                    self:FailEvent("Fail_Thaddius_Jump", destName)
                end
            end
        end
    end

    -- Light Bomb
    if spellId == 65120 then
        self.LastLight[destName] = timestamp
    end
end


function lib:SPELL_SUMMON(timestamp, type, sourceGUID, sourceName, sourceFlags, destGUID, destName, destFlags, spellId, spellName)
    --SPELL_SUMMON,0xF15000838601084D,"Aerial Command Unit",0xa48,0xF13000842C010A24,"Bomb Bot",0xa28,63811,"Bomb Bot",0x1
    if spellId == 63811 then
        -- could add the GUID for each bot if this doesnt work.
        self.SquelchFail.Fail_Mimiron_BombBots = timestamp + 30
        return
    end
end
        



-- 28801 Slime (Under Thaddius)
function lib:SPELL_AURA_REMOVED(timestamp, type, sourceGUID, sourceName, sourceFlags, destGUID, destName, destFlags, spellId, spellName, spellSchool)
    if spellId == 28801 then
        if (timestamp - self.DeathTime) < self.THADDIUS_JUMP_WINDOW then --error
            if bit.band(destFlags, COMBATLOG_OBJECT_TYPE_PLAYER) > 0 then
                self.LastSlime[destName] = timestamp
            end
        end
    end

    -- Light Bomb
    if spellId == 65120 then
        self.LastLight[destName] = nil
    end

    -- Gravity Bomb
    if spellId == 64233 then
        self.LastGravity[destName] = timestamp
    end
end



